@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python2
 # vim:fileencoding=utf-8
 from __future__ import (unicode_literals, division, absolute_import,
                         print_function)
@@ -6,7 +6,7 @@ from __future__ import (unicode_literals, division, absolute_import,
 __license__ = 'GPL v3'
 __copyright__ = '2013, Kovid Goyal <kovid at kovidgoyal.net>'
 
-import sys, os, re, math
+import sys, os, re, math, errno, uuid
 from collections import OrderedDict, defaultdict
 
 from lxml import html
@@ -15,9 +15,7 @@ from lxml.html.builder import (
 
 from calibre import guess_type
 from calibre.ebooks.docx.container import DOCX, fromstring
-from calibre.ebooks.docx.names import (
-    XPath, is_tag, XML, STYLES, NUMBERING, FONTS, get, generate_anchor,
-    descendants, FOOTNOTES, ENDNOTES, children, THEMES, SETTINGS)
+from calibre.ebooks.docx.names import XML, generate_anchor
 from calibre.ebooks.docx.styles import Styles, inherit, PageProperties
 from calibre.ebooks.docx.numbering import Numbering
 from calibre.ebooks.docx.fonts import Fonts
@@ -43,24 +41,34 @@ class Text:
         setattr(self.elem, self.attr, ''.join(self.buf))
         self.elem, self.attr, self.buf = elem, 'tail', []
 
+def html_lang(docx_lang):
+    lang = canonicalize_lang(docx_lang)
+    if lang and lang != 'und':
+        lang = lang_as_iso639_1(lang)
+        if lang:
+            return lang
+
 class Convert(object):
 
-    def __init__(self, path_or_stream, dest_dir=None, log=None, detect_cover=True, notes_text=None):
+    def __init__(self, path_or_stream, dest_dir=None, log=None, detect_cover=True, notes_text=None, notes_nopb=False, nosupsub=False):
         self.docx = DOCX(path_or_stream, log=log)
+        self.namespace = self.docx.namespace
         self.ms_pat = re.compile(r'\s{2,}')
         self.ws_pat = re.compile(r'[\n\r\t]')
         self.log = self.docx.log
         self.detect_cover = detect_cover
         self.notes_text = notes_text or _('Notes')
+        self.notes_nopb = notes_nopb
+        self.nosupsub = nosupsub
         self.dest_dir = dest_dir or os.getcwdu()
         self.mi = self.docx.metadata
         self.body = BODY()
-        self.theme = Theme()
-        self.settings = Settings()
-        self.tables = Tables()
-        self.fields = Fields()
-        self.styles = Styles(self.tables)
-        self.images = Images(self.log)
+        self.theme = Theme(self.namespace)
+        self.settings = Settings(self.namespace)
+        self.tables = Tables(self.namespace)
+        self.fields = Fields(self.namespace)
+        self.styles = Styles(self.namespace, self.tables)
+        self.images = Images(self.namespace, self.log)
         self.object_map = OrderedDict()
         self.html = HTML(
             HEAD(
@@ -77,11 +85,12 @@ class Convert(object):
             child.tail = '\n\t\t'
         self.html[0][-1].tail = '\n\t'
         self.html[1].text = self.html[1].tail = '\n'
-        lang = canonicalize_lang(self.mi.language)
-        if lang and lang != 'und':
-            lang = lang_as_iso639_1(lang)
-            if lang:
-                self.html.set('lang', lang)
+        lang = html_lang(self.mi.language)
+        if lang:
+            self.html.set('lang', lang)
+            self.doc_lang = lang
+        else:
+            self.doc_lang = None
 
     def __call__(self):
         doc = self.docx.document
@@ -91,10 +100,13 @@ class Convert(object):
         self.images(relationships_by_id)
         self.layers = OrderedDict()
         self.framed = [[]]
+        self.frame_map = {}
         self.framed_map = {}
         self.anchor_map = {}
         self.link_map = defaultdict(list)
         self.link_source_map = {}
+        self.toc_anchor = None
+        self.block_runs = []
         paras = []
 
         self.log.debug('Converting Word markup to HTML')
@@ -110,6 +122,7 @@ class Convert(object):
 
         self.read_block_anchors(doc)
         self.styles.apply_contextual_spacing(paras)
+        self.mark_block_runs(paras)
         # Apply page breaks at the start of every section, except the first
         # section (since that will be the start of the file)
         self.styles.apply_section_page_breaks(self.section_starts[1:])
@@ -138,6 +151,7 @@ class Convert(object):
                         dl[-1].append(p)
                         paras.append(wp)
                 self.styles.apply_contextual_spacing(paras)
+                self.mark_block_runs(paras)
 
         for p, wp in self.object_map.iteritems():
             if len(p) > 0 and not p.text and len(p[0]) > 0 and not p[0].text and p[0][0].get('class', None) == 'tab':
@@ -203,7 +217,7 @@ class Convert(object):
                 html_obj.set('class', cls)
 
         if notes_header is not None:
-            for h in children(self.body, 'h1', 'h2', 'h3'):
+            for h in self.namespace.children(self.body, 'h1', 'h2', 'h3'):
                 notes_header.tag = h.tag
                 cls = h.get('class', None)
                 if cls and cls != 'notes-header':
@@ -213,7 +227,7 @@ class Convert(object):
         self.fields.polish_markup(self.object_map)
 
         self.log.debug('Cleaning up redundant markup generated by Word')
-        self.cover_image = cleanup_markup(self.log, self.html, self.styles, self.dest_dir, self.detect_cover)
+        self.cover_image = cleanup_markup(self.log, self.html, self.styles, self.dest_dir, self.detect_cover, self.namespace.XPath)
 
         return self.write(doc)
 
@@ -222,14 +236,14 @@ class Convert(object):
         self.page_map = OrderedDict()
         self.section_starts = []
 
-        for p in descendants(doc, 'w:p', 'w:tbl'):
+        for p in self.namespace.descendants(doc, 'w:p', 'w:tbl'):
             if p.tag.endswith('}tbl'):
                 self.tables.register(p, self.styles)
                 current.append(p)
                 continue
-            sect = tuple(descendants(p, 'w:sectPr'))
+            sect = tuple(self.namespace.descendants(p, 'w:sectPr'))
             if sect:
-                pr = PageProperties(sect)
+                pr = PageProperties(self.namespace, sect)
                 paras = current + [p]
                 for x in paras:
                     self.page_map[x] = pr
@@ -240,8 +254,8 @@ class Convert(object):
 
         if current:
             self.section_starts.append(current[0])
-            last = XPath('./w:body/w:sectPr')(doc)
-            pr = PageProperties(last)
+            last = self.namespace.XPath('./w:body/w:sectPr')(doc)
+            pr = PageProperties(self.namespace, last)
             for x in current:
                 self.page_map[x] = pr
 
@@ -256,16 +270,16 @@ class Convert(object):
                     name = name
             return name
 
-        nname = get_name(NUMBERING, 'numbering.xml')
-        sname = get_name(STYLES, 'styles.xml')
-        sename = get_name(SETTINGS, 'settings.xml')
-        fname = get_name(FONTS, 'fontTable.xml')
-        tname = get_name(THEMES, 'theme1.xml')
-        foname = get_name(FOOTNOTES, 'footnotes.xml')
-        enname = get_name(ENDNOTES, 'endnotes.xml')
-        numbering = self.numbering = Numbering()
-        footnotes = self.footnotes = Footnotes()
-        fonts = self.fonts = Fonts()
+        nname = get_name(self.namespace.names['NUMBERING'], 'numbering.xml')
+        sname = get_name(self.namespace.names['STYLES'], 'styles.xml')
+        sename = get_name(self.namespace.names['SETTINGS'], 'settings.xml')
+        fname = get_name(self.namespace.names['FONTS'], 'fontTable.xml')
+        tname = get_name(self.namespace.names['THEMES'], 'theme1.xml')
+        foname = get_name(self.namespace.names['FOOTNOTES'], 'footnotes.xml')
+        enname = get_name(self.namespace.names['ENDNOTES'], 'endnotes.xml')
+        numbering = self.numbering = Numbering(self.namespace)
+        footnotes = self.footnotes = Footnotes(self.namespace)
+        fonts = self.fonts = Fonts(self.namespace)
 
         foraw = enraw = None
         forel, enrel = ({}, {}), ({}, {})
@@ -274,6 +288,10 @@ class Convert(object):
                 seraw = self.docx.read(sename)
             except KeyError:
                 self.log.warn('Settings %s do not exist' % sename)
+            except EnvironmentError as e:
+                if e.errno != errno.ENOENT:
+                    raise
+                self.log.warn('Settings %s file missing' % sename)
             else:
                 self.settings(fromstring(seraw))
 
@@ -329,11 +347,11 @@ class Convert(object):
         self.styles.resolve_numbering(numbering)
 
     def write(self, doc):
-        toc = create_toc(doc, self.body, self.resolved_link_map, self.styles, self.object_map, self.log)
+        toc = create_toc(doc, self.body, self.resolved_link_map, self.styles, self.object_map, self.log, self.namespace)
         raw = html.tostring(self.html, encoding='utf-8', doctype='<!DOCTYPE html>')
         with open(os.path.join(self.dest_dir, 'index.html'), 'wb') as f:
             f.write(raw)
-        css = self.styles.generate_css(self.dest_dir, self.docx)
+        css = self.styles.generate_css(self.dest_dir, self.docx, self.notes_nopb, self.nosupsub)
         if css:
             with open(os.path.join(self.dest_dir, 'docx.css'), 'wb') as f:
                 f.write(css.encode('utf-8'))
@@ -347,19 +365,23 @@ class Convert(object):
         opf.create_spine(['index.html'])
         if self.cover_image is not None:
             opf.guide.set_cover(self.cover_image)
+        def process_guide(E, guide):
+            if self.toc_anchor is not None:
+                guide.append(E.reference(
+                    href='index.html#' + self.toc_anchor, title=_('Table of Contents'), type='toc'))
         toc_file = os.path.join(self.dest_dir, 'toc.ncx')
         with open(os.path.join(self.dest_dir, 'metadata.opf'), 'wb') as of, open(toc_file, 'wb') as ncx:
-            opf.render(of, ncx, 'toc.ncx')
+            opf.render(of, ncx, 'toc.ncx', process_guide=process_guide)
         if os.path.getsize(toc_file) == 0:
             os.remove(toc_file)
         return os.path.join(self.dest_dir, 'metadata.opf')
 
     def read_block_anchors(self, doc):
-        doc_anchors = frozenset(XPath('./w:body/w:bookmarkStart[@w:name]')(doc))
+        doc_anchors = frozenset(self.namespace.XPath('./w:body/w:bookmarkStart[@w:name]')(doc))
         if doc_anchors:
             current_bm = set()
             rmap = {v:k for k, v in self.object_map.iteritems()}
-            for p in descendants(doc, 'w:p', 'w:bookmarkStart[@w:name]'):
+            for p in self.namespace.descendants(doc, 'w:p', 'w:bookmarkStart[@w:name]'):
                 if p.tag.endswith('}p'):
                     if current_bm and p in rmap:
                         para = rmap[p]
@@ -369,7 +391,7 @@ class Convert(object):
                             self.anchor_map[name] = para.get('id')
                         current_bm = set()
                 elif p in doc_anchors:
-                    anchor = get(p, 'w:name')
+                    anchor = self.namespace.get(p, 'w:name')
                     if anchor:
                         current_bm.add(anchor)
 
@@ -378,11 +400,12 @@ class Convert(object):
         self.object_map[dest] = p
         style = self.styles.resolve_paragraph(p)
         self.layers[p] = []
+        self.frame_map[p] = style.frame
         self.add_frame(dest, style.frame)
 
         current_anchor = None
         current_hyperlink = None
-        hl_xpath = XPath('ancestor::w:hyperlink[1]')
+        hl_xpath = self.namespace.XPath('ancestor::w:hyperlink[1]')
 
         def p_parent(x):
             # Ensure that nested <w:p> tags are handled. These can occur if a
@@ -395,7 +418,7 @@ class Convert(object):
                 except AttributeError:
                     break
 
-        for x in descendants(p, 'w:r', 'w:bookmarkStart', 'w:hyperlink'):
+        for x in self.namespace.descendants(p, 'w:r', 'w:bookmarkStart', 'w:hyperlink', 'w:instrText'):
             if p_parent(x) is not p:
                 continue
             if x.tag.endswith('}r'):
@@ -414,8 +437,10 @@ class Convert(object):
                 dest.append(span)
                 self.layers[p].append(x)
             elif x.tag.endswith('}bookmarkStart'):
-                anchor = get(x, 'w:name')
-                if anchor and anchor not in self.anchor_map:
+                anchor = self.namespace.get(x, 'w:name')
+                if anchor and anchor not in self.anchor_map and anchor != '_GoBack':
+                    # _GoBack is a special bookmark inserted by Word 2010 for
+                    # the return to previous edit feature, we ignore it
                     old_anchor = current_anchor
                     self.anchor_map[anchor] = current_anchor = generate_anchor(anchor, frozenset(self.anchor_map.itervalues()))
                     if old_anchor is not None:
@@ -425,6 +450,16 @@ class Convert(object):
                                 self.anchor_map[a] = current_anchor
             elif x.tag.endswith('}hyperlink'):
                 current_hyperlink = x
+            elif x.tag.endswith('}instrText') and x.text and x.text.strip().startswith('TOC '):
+                old_anchor = current_anchor
+                anchor = str(uuid.uuid4())
+                self.anchor_map[anchor] = current_anchor = generate_anchor('toc', frozenset(self.anchor_map.itervalues()))
+                self.toc_anchor = current_anchor
+                if old_anchor is not None:
+                    # The previous anchor was not applied to any element
+                    for a, t in tuple(self.anchor_map.iteritems()):
+                        if t == old_anchor:
+                            self.anchor_map[a] = current_anchor
 
         m = re.match(r'heading\s+(\d+)$', style.style_name or '', re.IGNORECASE)
         if m is not None:
@@ -458,7 +493,7 @@ class Convert(object):
                 wrapper = self.wrap_elems(spans, SPAN())
                 wrapper.set('class', cls)
 
-        if not dest.text and len(dest) == 0:
+        if not dest.text and len(dest) == 0 and not style.has_visible_border():
             # Empty paragraph add a non-breaking space so that it is rendered
             # by WebKit
             dest.text = NBSP
@@ -481,7 +516,13 @@ class Convert(object):
         wrapper.tail = elems[-1].tail
         elems[-1].tail = None
         for elem in elems:
-            p.remove(elem)
+            try:
+                p.remove(elem)
+            except ValueError:
+                # Probably a hyperlink that spans multiple
+                # paragraphs,theoretically we should break this up into
+                # multiple hyperlinks, but I can't be bothered.
+                elem.getparent().remove(elem)
             wrapper.append(elem)
         return wrapper
 
@@ -494,17 +535,17 @@ class Convert(object):
                 span = self.wrap_elems(spans, SPAN())
             span.tag = 'a'
             self.resolved_link_map[hyperlink] = span
-            tgt = get(hyperlink, 'w:tgtFrame')
+            tgt = self.namespace.get(hyperlink, 'w:tgtFrame')
             if tgt:
                 span.set('target', tgt)
-            tt = get(hyperlink, 'w:tooltip')
+            tt = self.namespace.get(hyperlink, 'w:tooltip')
             if tt:
                 span.set('title', tt)
-            rid = get(hyperlink, 'r:id')
+            rid = self.namespace.get(hyperlink, 'r:id')
             if rid and rid in relationships_by_id:
                 span.set('href', relationships_by_id[rid])
                 continue
-            anchor = get(hyperlink, 'w:anchor')
+            anchor = self.namespace.get(hyperlink, 'w:anchor')
             if anchor and anchor in self.anchor_map:
                 span.set('href', '#' + self.anchor_map[anchor])
                 continue
@@ -568,27 +609,31 @@ class Convert(object):
         text = Text(ans, 'text', [])
 
         for child in run:
-            if is_tag(child, 'w:t'):
+            if self.namespace.is_tag(child, 'w:t'):
                 if not child.text:
                     continue
                 space = child.get(XML('space'), None)
                 preserve = False
-                if space == 'preserve':
-                    # Only use a <span> with white-space:pre-wrap if this element
-                    # actually needs it, i.e. if it has more than one
-                    # consecutive space or it has newlines or tabs.
-                    multi_spaces = self.ms_pat.search(child.text) is not None
-                    preserve = multi_spaces or self.ws_pat.search(child.text) is not None
+                ctext = child.text
+                if space != 'preserve':
+                    # Remove leading and trailing whitespace. Word ignores
+                    # leading and trailing whitespace without preserve
+                    ctext = ctext.strip(' \n\r\t')
+                # Only use a <span> with white-space:pre-wrap if this element
+                # actually needs it, i.e. if it has more than one
+                # consecutive space or it has newlines or tabs.
+                multi_spaces = self.ms_pat.search(ctext) is not None
+                preserve = multi_spaces or self.ws_pat.search(ctext) is not None
                 if preserve:
-                    text.add_elem(SPAN(child.text, style="white-space:pre-wrap"))
+                    text.add_elem(SPAN(ctext, style="white-space:pre-wrap"))
                     ans.append(text.elem)
                 else:
-                    text.buf.append(child.text)
-            elif is_tag(child, 'w:cr'):
+                    text.buf.append(ctext)
+            elif self.namespace.is_tag(child, 'w:cr'):
                 text.add_elem(BR())
                 ans.append(text.elem)
-            elif is_tag(child, 'w:br'):
-                typ = get(child, 'w:type')
+            elif self.namespace.is_tag(child, 'w:br'):
+                typ = self.namespace.get(child, 'w:type')
                 if typ in {'column', 'page'}:
                     br = BR(style='page-break-after:always')
                 else:
@@ -599,25 +644,25 @@ class Convert(object):
                         br = BR()
                 text.add_elem(br)
                 ans.append(text.elem)
-            elif is_tag(child, 'w:drawing') or is_tag(child, 'w:pict'):
+            elif self.namespace.is_tag(child, 'w:drawing') or self.namespace.is_tag(child, 'w:pict'):
                 for img in self.images.to_html(child, self.current_page, self.docx, self.dest_dir):
                     text.add_elem(img)
                     ans.append(text.elem)
-            elif is_tag(child, 'w:footnoteReference') or is_tag(child, 'w:endnoteReference'):
+            elif self.namespace.is_tag(child, 'w:footnoteReference') or self.namespace.is_tag(child, 'w:endnoteReference'):
                 anchor, name = self.footnotes.get_ref(child)
                 if anchor and name:
-                    l = SUP(A(name, href='#' + anchor, title=name), id='back_%s' % anchor)
+                    l = A(SUP(name, id='back_%s' % anchor), href='#' + anchor, title=name)
                     l.set('class', 'noteref')
                     text.add_elem(l)
                     ans.append(text.elem)
-            elif is_tag(child, 'w:tab'):
+            elif self.namespace.is_tag(child, 'w:tab'):
                 spaces = int(math.ceil((self.settings.default_tab_stop / 36) * 6))
                 text.add_elem(SPAN(NBSP * spaces))
                 ans.append(text.elem)
                 ans[-1].set('class', 'tab')
-            elif is_tag(child, 'w:noBreakHyphen'):
+            elif self.namespace.is_tag(child, 'w:noBreakHyphen'):
                 text.buf.append(u'\u2011')
-            elif is_tag(child, 'w:softHyphen'):
+            elif self.namespace.is_tag(child, 'w:softHyphen'):
                 text.buf.append(u'\u00ad')
         if text.buf:
             setattr(text.elem, text.attr, ''.join(text.buf))
@@ -626,7 +671,9 @@ class Convert(object):
         if style.vert_align in {'superscript', 'subscript'}:
             ans.tag = 'sub' if style.vert_align == 'subscript' else 'sup'
         if style.lang is not inherit:
-            ans.lang = style.lang
+            lang = html_lang(style.lang)
+            if lang is not None and lang != self.doc_lang:
+                ans.set('lang', lang)
         return ans
 
     def add_frame(self, html_obj, style):
@@ -655,6 +702,62 @@ class Convert(object):
             self.framed_map[frame] = css = style.css(self.page_map[self.object_map[paras[0]]])
             self.styles.register(css, 'frame')
 
+        if not self.block_runs:
+            return
+        rmap = {v:k for k, v in self.object_map.iteritems()}
+        for border_style, blocks in self.block_runs:
+            paras = tuple(rmap[p] for p in blocks)
+            parent = paras[0].getparent()
+            idx = parent.index(paras[0])
+            frame = DIV(*paras)
+            parent.insert(idx, frame)
+            self.framed_map[frame] = css = border_style.css
+            self.styles.register(css, 'frame')
+
+    def mark_block_runs(self, paras):
+
+        def process_run(run):
+            max_left = max_right = 0
+            has_visible_border = None
+            for p in run:
+                style = self.styles.resolve_paragraph(p)
+                if has_visible_border is None:
+                    has_visible_border = style.has_visible_border()
+                max_left, max_right = max(style.margin_left, max_left), max(style.margin_right, max_right)
+                if has_visible_border:
+                    style.margin_left = style.margin_right = inherit
+                if p is not run[0]:
+                    style.padding_top = 0
+                else:
+                    border_style = style.clone_border_styles()
+                    if has_visible_border:
+                        border_style.margin_top, style.margin_top = style.margin_top, inherit
+                if p is not run[-1]:
+                    style.padding_bottom = 0
+                else:
+                    if has_visible_border:
+                        border_style.margin_bottom, style.margin_bottom = style.margin_bottom, inherit
+                style.clear_borders()
+                if p is not run[-1]:
+                    style.apply_between_border()
+            if has_visible_border:
+                border_style.margin_left, border_style.margin_right = max_left,max_right
+                self.block_runs.append((border_style, run))
+
+        run = []
+        for p in paras:
+            if run and self.frame_map.get(p) == self.frame_map.get(run[-1]):
+                style = self.styles.resolve_paragraph(p)
+                last_style = self.styles.resolve_paragraph(run[-1])
+                if style.has_identical_borders(last_style):
+                    run.append(p)
+                    continue
+            if len(run) > 1:
+                process_run(run)
+            run = [p]
+        if len(run) > 1:
+            process_run(run)
+
 if __name__ == '__main__':
     import shutil
     from calibre.utils.logging import default_log
